#include <fstream>
#include <iostream>
#include <vector>
#include <cstring>
#include <sys/stat.h>
#include <stdint.h>

#define NAME_LENGTH 56

using namespace std;

struct PakEntry
{
	public:
		char	file[NAME_LENGTH];
		int32_t	position;
		int32_t	length;
};

struct PakHeader
{
	public:
		char	header[4];
		int32_t	position;
		int32_t	length;
};

string &strtolower(string &str)
{
	for(int i = 0;i < str.length();i++)
	{
		if(str[i] >= 'A' && str[i] <= 'Z')
			str.replace(i, 1, 1, str[i] + ('a'-'A'));
	}
	return str;
}

int main(int argc, char* argv[])
{
	cout << "Heretic 2 Data Extracter\n\n";
	if(argc < 2)
	{
		cout << "	unpak <out_dir> <file...>\n";
		return 0;
	}

	for(int f = 2;f < argc;f++)
	{
		fstream pak(argv[f], ios_base::in|ios_base::binary);
		if(!pak.is_open())
		{
			cout << "ERROR: Could not open \"" << argv[f] << "\"!\n";
			continue;
		}

		PakHeader header;
		pak.read((char*) &header, sizeof(PakHeader));
		if(header.header[0] != 'P' && header.header[1] != 'A' && header.header[2] != 'C' && header.header[3] != 'K')
		{
			cout << "ERROR: \"" << argv[f] << "\" is not a valid pak file!\n";
			pak.close();
			continue;
		}
		const int numFiles = header.length/sizeof(PakEntry);
		cout << "\"" << argv[f] << "\" has " << numFiles << " files.\n";

		if(pak.seekg(header.position, ios_base::beg).fail())
		{
			cout << "ERROR: While Reading \"" << argv[f] << "\"!\n";
			pak.close();
			continue;
		}
		vector<PakEntry> files;
		PakEntry file;
		for(int i = 0;i < numFiles;i++)
		{
			file = PakEntry();
			pak.read((char*) &file, sizeof(PakEntry));
			files.push_back(file);
		}

		bool undo = false;
		for(int i = 0;i < numFiles;i++)
		{
			if(i != 0 && undo)
			{
				cout << string(strlen(files[i-1].file) + 4, '\b');
				cout << string(strlen(files[i-1].file) + 4, ' ');
				cout << string(strlen(files[i-1].file) + 4, '\b');
			}
			else
				cout << "Extracting: \"";
			undo = true;
			string filename(files[i].file);
			strtolower(filename);
			cout << filename << "\"...";
			filename = string(argv[1]).append("/").append(filename);

			if(pak.seekg(files[i].position, ios_base::beg).fail())
			{
				cout << "\nERROR: Could not extract \"" << files[i].file << "\"!\n";
				undo = false;
				continue;
			}
			char* data = new char[files[i].length];
			pak.read(data, files[i].length);

			for(size_t j = filename.find_first_of('/');j != string::npos;j = filename.find_first_of('/', j+1))
			{
				mkdir(filename.substr(0, j).c_str(), 0777);
			}
			fstream out(filename.c_str(), ios_base::out|ios_base::trunc);
			if(!out.is_open())
			{
				cout << "\nError: Could not write to \"" << filename << "\"!\n";
				undo = false;
				out.close();
				continue;
			}
			out.write(data, files[i].length);
			out.close();

			delete[] data;
		}
		cout << "\n";

		pak.close();
		cout << "Done!\n";
	}
}
